/*
 * TextSuite Java Wrapper by Markus Dllinger
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * 
 * TextSuite (C) Steffen Xonna (aka Lossy eX)
 * http://www.opengl24.de/
 * 
 * -----------------------------------------------------
 * For copyright information, see the file copyright.txt
 * 
 */

package textsuite.samples;
import java.awt.Dimension;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.IntBuffer;

import javax.swing.JFrame;

import org.lwjgl.LWJGLException;
import org.lwjgl.opengl.AWTGLCanvas;
import org.lwjgl.opengl.GL11;

import textsuite.BufferUtils;
import textsuite.IPostProcessProc;
import textsuite.ImageBlurArguments;
import textsuite.TextSuite;

public class TextSuite_LWJGL extends JFrame {

	private static final long serialVersionUID = -4793715083754138263L;
	private AWTOpenGLCanvas canvas;

	public static void main( String[] args ) throws LWJGLException, InterruptedException, IOException 
	{ 
		// Create a window
		TextSuite_LWJGL win = new TextSuite_LWJGL(); 
		win.setVisible(true); 
		
		// Enter the main loop
		win.canvas.mainLoop();

		// Dispose everything
		win.dispose();
	}

	public TextSuite_LWJGL() throws LWJGLException, InterruptedException {

		// Add a window adapter in order to react to the quit message
		WindowListener listener = new WindowAdapter() {
			public void windowClosing(WindowEvent evt) {
				canvas.cancel = true;
			}
		};

		addWindowListener(listener);
		setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE); 

		// Create a Canvas. Note that this automatically creates an OpenGL context
		canvas = new AWTOpenGLCanvas();
		canvas.setSize(new Dimension(800, 600));
		add(canvas);
		this.pack();
	}

}

class AWTOpenGLCanvas extends AWTGLCanvas {
	
	private int contextID;
	private int fontID1;
	private int fontID2;

	private static final long serialVersionUID = -5030829228222844560L;
	
	
	/**
	 * Indicates whether the cleanup method should be called
	 */
	public boolean cancel = false;
	
	/**
	 * Indicates that the main loop should end
	 */
	private boolean doCancel = false;
	
	
	/**
	 * Indicates that the viewport changed
	 */
	boolean doResize = false;
	

	public AWTOpenGLCanvas() throws LWJGLException {
		super();
	}

	public void mainLoop() {
		while (!doCancel)
			repaint();
	}


	
	public void checkForErrors() {
		int error = TextSuite.tsGetError();
		if (error != TextSuite.TS_NO_ERROR)
			System.out.println("Error: " + TextSuite.tsGetErrorString(error));
	}
	
	public void cleanup() {
		TextSuite.tsFontDestroy(fontID1);
		TextSuite.tsFontDestroy(fontID2);
		TextSuite.tsContextDestroy(contextID);
		TextSuite.tsQuit();
	}
	
	public void initLogic() {
		addComponentListener(new ComponentAdapter() {

			@Override
			public void componentResized(ComponentEvent e) {
				doResize = true;
			}
		});
	}
	
	protected void initGL() {
		
		// Initialize the logic
		initLogic();

		// Set some OpenGL parameters
		GL11.glClearColor(0.4f, 0.4f, 0.4f, 1.0f);

		GL11.glMatrixMode(GL11.GL_PROJECTION);
		GL11.glLoadIdentity();		
		GL11.glOrtho(0, 800, 600, 0, 0.0, 1024);
		GL11.glMatrixMode(GL11.GL_MODELVIEW);
		GL11.glHint(GL11.GL_PERSPECTIVE_CORRECTION_HINT, GL11.GL_NICEST); 

		GL11.glEnable(GL11.GL_TEXTURE_2D);
		GL11.glBlendFunc(GL11.GL_SRC_ALPHA, GL11.GL_ONE_MINUS_SRC_ALPHA);
		GL11.glEnable(GL11.GL_BLEND);


		// Initialize TextSuite
		if (!TextSuite.tsInit(TextSuite.TS_INIT_OPENGL | 
				TextSuite.TS_INIT_TEXTSUITE | 
				TextSuite.TS_INIT_GDI))
			throw new RuntimeException("Could not initialize TextSuite!");
		

		// Create a context
		contextID = TextSuite.tsContextCreate();
		TextSuite.tsContextBind(contextID);
		
		// Check for errors
		checkForErrors();

		// Set some parameters
		TextSuite.tsSetParameteri(TextSuite.TS_RENDERER, TextSuite.TS_RENDERER_OPENGL);
		TextSuite.tsSetParameteri(TextSuite.TS_CREATOR, TextSuite.TS_CREATOR_GDI_FACENAME);

		// Create a new font
		fontID1 = TextSuite.tsFontCreateCreatorA("Arial Black", 30, TextSuite.TS_STYLE_NORMAL, TextSuite.TS_DEFAULT, TextSuite.TS_DEFAULT);
		TextSuite.tsFontBind(fontID1);
		
		checkForErrors();
		
		// Add a custom post-processor callback
		//
		//	Note that we could define user data by setting the second argument
		//	to some Object. We could access it in the callback via the argument 'data'.
		TextSuite.tsPostAddCustom(new IPostProcessProc() {

			@Override
			public void execute(int ImageID, char CharCode, Object data) {

				int TempImageID = TextSuite.tsImageCreate();
				TextSuite.tsImageAssignFrom(TempImageID, ImageID);

				// fill with color
				TextSuite.tsImageFillColor3ub(TempImageID, 0xE0, 0xE0, 0xFF, TextSuite.TS_CHANNELS_RGB);

				// original image will be the shadow so blur image
				ImageBlurArguments res = TextSuite.tsImageBlur(ImageID, 2, 8, TextSuite.TS_CHANNEL_ALPHA, true);

				// fill with color
				TextSuite.tsImageFillColor3ub(ImageID, 0x00, 0x00, 0xFF, TextSuite.TS_CHANNELS_RGB);

				// blend cloned original image over blurred original
				TextSuite.tsImageBlend(ImageID, TempImageID, res.expandSizeX, res.expandSizeY + 5, true);

				// destroy temp
				TextSuite.tsImageDestroy(TempImageID);

				// the glyph rect has changed so get it
				IntBuffer buf = BufferUtils.createIntBuffer(4);
				TextSuite.tsFontGetCharParameteriv(CharCode, TextSuite.TS_CHAR_GLYPHRECT, buf);

				int left = buf.get() + res.expandSizeX;
				int top = buf.get() + res.expandSizeY + 5;
				int right = buf.get() + res.expandSizeX;
				int bottom = buf.get() + res.expandSizeY + 5;

				buf.flip();
				buf.put(left);
				buf.put(top);
				buf.put(right);
				buf.put(bottom);

				// and set it back
				TextSuite.tsFontSetCharParameteriv(CharCode, TextSuite.TS_CHAR_GLYPHRECT, buf);
			}
		}, null);
		
		checkForErrors();
		
		// Create a second font
		fontID2 = TextSuite.tsFontCreateCreatorA("Lucida", 30, TextSuite.TS_STYLE_NORMAL, TextSuite.TS_DEFAULT, TextSuite.TS_DEFAULT);
		TextSuite.tsFontBind(fontID2);
		
		checkForErrors();

		// fill with orange
		TextSuite.tsPostAddFillColor3ub(0xE0, 0x80, 0x20, TextSuite.TS_CHANNELS_RGB);

		// white border
		TextSuite.tsPostAddBorder3ub(1.5f, 1f, 0xFF, 0xFF, 0xFF);
		TextSuite.tsPostAddUsageRange(TextSuite.TS_POST_INDEX_LAST, TextSuite.TS_POST_USAGE_INCLUDE, 'A', 'Z');
		TextSuite.tsPostAddShadow4ub(2f, 0, 0, 0xFF, 0xFF, 0xFF, 0x80);

		// shadow
		TextSuite.tsPostAddShadow3ub(2, 0, 0, 0x30, 0x30, 0x30);
		TextSuite.tsPostAddShadow3ub(5, 0, 0, 0x30, 0x30, 0x30);

		// Add a wave post-processor
		TextSuite.tsPostAddCustom(new IPostProcessProc() {

			@Override
			public void execute(int ImageID, char CharCode, Object data) {
				
				// Strength of the waves
				final int WIDE = 2;
				
				// Expand image
				int ImageHeight = TextSuite.tsImageGetHeight(ImageID);
				int ImageWidth = TextSuite.tsImageGetWidth(ImageID);
				
				int BaseLine = TextSuite.tsFontGetCharParameteri(CharCode, TextSuite.TS_CHAR_GLYPHORIGIN_Y);

				TextSuite.tsImageResize(ImageID, ImageWidth + WIDE*2, ImageHeight, WIDE, 0);
				
				// Move the pixels in each line according to the offset
				for (int y = 0; y < ImageHeight; y++) {
					
					// Compute the wave offset
					int offset = (int) Math.round((float) Math.sin((y - BaseLine) / 2.5f) * WIDE);
					
					// Get the scanline buffer of the current line.
					//
					// 	Note that the buffer contains the entire image data,
					//  only the position is set to the beginning of the line.
					//
					//	Changes to this buffer will directly affect the image
					//	data in the TextSuite library.
					ByteBuffer pix = TextSuite.tsImageScanline(ImageID, y);
					ByteBuffer pix2 = null;
					
					// Multiply the offset with 4 in order to skip a single pixel (RGBA)
					pix.position(pix.position() + WIDE * 4);
					
					if (offset > 0) {
						pix.position(pix.position() + ImageWidth * 4);
						pix2 = pix.duplicate();
						
						// avoid buffer overflow in case the offset is 2 in the last row
						if (y == (ImageHeight - 1))
							offset = Math.round(offset / 2);
						
						pix2.position(pix2.position() + offset * 4);


						for (int x = 0; x <= ImageWidth + WIDE; x++) {
							// read/write 4 bytes (RGBA)
							pix2.put(pix.get());
							pix2.put(pix.get());
							pix2.put(pix.get());
							pix2.put(pix.get());
							if (x < ImageWidth + WIDE) {
								pix2.position(pix2.position() - 2 * 4);
								pix.position(pix.position() - 2 * 4);
							}
						}
					} else if (offset < 0) {
						pix.position(pix.position() - WIDE * 4);
						pix2 = pix.duplicate();
						pix2.position(pix2.position() - offset * 4);

						for (int x = 0; x < ImageWidth + WIDE; x++) {
							pix.put(pix2.get());
							pix.put(pix2.get());
							pix.put(pix2.get());
							pix.put(pix2.get());
						}
					}
				}

				// Update the parameters
				TextSuite.tsFontSetCharParameteri(CharCode, TextSuite.TS_CHAR_GLYPHRECT_LEFT, 
						TextSuite.tsFontGetCharParameteri(CharCode, TextSuite.TS_CHAR_GLYPHRECT_LEFT) + WIDE);
				TextSuite.tsFontSetCharParameteri(CharCode, TextSuite.TS_CHAR_GLYPHRECT_RIGHT, 
						TextSuite.tsFontGetCharParameteri(CharCode, TextSuite.TS_CHAR_GLYPHRECT_RIGHT) + WIDE);

			}
		}, null);


	}

	protected void paintGL() {
		
		// cleanup and set the doCancel flag
		if (cancel) {
			cleanup();
			doCancel = true;
			cancel = false;
			return;
		}
		
		// Perform some OpenGL tasks
		GL11.glClear(GL11.GL_COLOR_BUFFER_BIT | GL11.GL_STENCIL_BUFFER_BIT);
		GL11.glMatrixMode(GL11.GL_MODELVIEW);
		GL11.glLoadIdentity();
		
		// Resize the viewport, if necessary
		if (doResize) {
			GL11.glMatrixMode(GL11.GL_PROJECTION);
			GL11.glLoadIdentity();		
			GL11.glViewport(0, 0, this.getWidth(), this.getHeight());
			GL11.glOrtho(0, this.getWidth(), this.getHeight(), 0, 0.0, 1024);
			GL11.glMatrixMode(GL11.GL_MODELVIEW);
			doResize = false;
		}
		
		// Align the text with TS_ALIGN_CENTER
		TextSuite.tsSetParameteri(TextSuite.TS_ALIGN, TextSuite.TS_ALIGN_CENTER);

		// Print the TextSuite copyright string
		String text = TextSuite.tsGetString(TextSuite.TS_INFO_COPYRIGHT);
		
		// Bind the first font
		TextSuite.tsFontBind(fontID1);
		GL11.glTranslatef(getWidth() / 2f, getHeight() / 4f + TextSuite.tsTextGetHeight(text) / 2f, 0.0f);
		TextSuite.tsTextOut(text);
		
		// Bind the second font
		TextSuite.tsFontBind(fontID2);
		text = "TextSuite Java Wrapper by Markus Dllinger";
		GL11.glLoadIdentity();
		GL11.glTranslatef(getWidth() / 2f, getHeight() * 3f / 4f - TextSuite.tsTextGetHeight(text) / 2f, 0.0f);
		TextSuite.tsTextOut(text);

		checkForErrors();

		// Swap the buffers
		try {
			swapBuffers();
		} catch (LWJGLException e) {
			e.printStackTrace();
		}
	}
}

